# About
This documentation is part of [Pod](../README.md).

GraphQL endpoint on HTTP API is Pods interface to fetch data for a subset of the [GraphQL](https://graphql.org/) [specification](https://spec.graphql.org/October2021/).
# API Authentication & Credentials
Accessing your Pod requires authentication, this works the same as [HTTP API](HTTP_API.md).

# GraphQL API

The GraphQL specification provides an efficient interface for fetching graph data that consist of nodes and edges. Pods `/graphql` endpoint allows you to query your data according to this specification. Since Pod currently only support a subset of all graphQL features, please carefully read this document first.

## Basic Data Retrieval
A basic GraphQL query allows for selecting a subset of item properties and edges in the form of a tree.
```json
{
  query {
    Person {
      firstName
      lastName
      birthDate
      address: {
        city
      }
    }
  }
}
```

In Pods implementation of graphQL, every query uses a set of items as a starting point for the query. We refer to this set as the "anchoring items". You can anchor your query by providing a type, in this case the `Person` type.

## Filters
Items can be filtered by a variaty of operators, namely eq, ne, gt, lt, gte, lte, and, or.
```json
{
  query {
    Account (filter: {dateCreated: {gte: 1654784703}}) {
        name
    }
  }
}
```

Filters can be applied to edges as well:
```json
{
  query {
    Person {
      ~owner (filter: {name: {eq: "Alice"}}) {
          name
      }
    }
  }
}
```

Filter can be combined using and/or logic. Note that every `and` or `or` clause can only have 2 children, but can be nested.
```json
{
  query {
    Person {
      owner (filter: {and: [{name: {neq: "Alice"}}, name: {neq: "Bob"}}]) {
          id
      }
    }
  }
}
```


## Reverse Edges
Pod's graphql API allows you to access reverse edges by prefixing edges with a '~' (tilde) character. An example query that fetches all items that have an edge `MyItem` `-owner->` `Person` is shown below:

```json
{
  query {
    Person {
      ~owner {
        id
        name
      }
    }
  }
}
```

## Pagination
Pagination is possible by using limit and offset variables in a query. 
```json
{
  query {
    Message (limit: 10, offset:5) {
      subject
      sender {
        name
      }
    }
  }
}
```

## Sorting
To sort items "order_desc" and "order_asc" arguments are available. 
```json
{
  query {
    Account (order_desc: dateCreated) {
      id
      name
    }
  }
}
```

## Aggregation (Coming soon)
GraphQL allows for aggregating results at each level. The `count` keyword can be used for the nr of items on the specified level in the result. The `total` keyword can be used for the total nr of items in the database. 
```json
{
  query {
    Message {
      subject
      service
    }
    count
    total
  }
}
```

Those GraphQL queries above are wrapped around to conform to the usual HTTP API object while interacting with Pod.
```json
{
  "auth": $auth_json,
  "payload": $grapql_query
}
```

## Example request and response

### Request
```json
{
  "auth": $auth_json,
  "payload": """{
    query {
      Message (limit: 2) {
        subject
        service
        sender (filter: {name: {eq: "memri"}}) {
          name
        }
      }
      count
      total
    }
  }"""
}
```

### Response
```json
{
  "data": [
    {"subject":"Hello memri", "service": "memri", "sender": []},
    {"subject":"Hello chat app", "service": "chatapp", "sender": [
      {"name": "memri"}
    ]}
  ],
  "count": 2,
  "total": 122
}

