use crate::api_model::Bulk;
use crate::api_model::CreateEdge;
use crate::api_model::CreateItem;
use crate::api_model::DeleteEdgeBySourceTarget;
use crate::api_model::EdgeDirection;
use crate::api_model::GetEdges;
use crate::api_model::Search;
use crate::api_model::SortOrder;
use crate::command_line_interface::CliOptions;
use crate::database_api;
use crate::database_api::get_incoming_edges;
use crate::database_api::get_outgoing_edges;
use crate::database_api::DatabaseSearch;
use crate::database_api::EdgePointer;
use crate::database_api::GQLSearchArgs;
use crate::database_api::Rowid;
use crate::database_utils::add_item_edge_properties;
use crate::database_utils::insert_edge;
use crate::database_utils::insert_property;
use crate::database_utils::item_base_to_json;
use crate::error::Error;
use crate::error::Result;
use crate::graphql_utils;
use crate::graphql_utils::QueryASTNode;
use crate::plugin_auth_crypto::DatabaseKey;
use crate::schema;
use crate::schema::validate_property_name;
use crate::schema::Schema;
use crate::triggers;
use crate::triggers::SchemaAdditionChange;
use log::info;
use rand::Rng;
use rusqlite::Transaction as Tx;
use serde_json::Value;
use std::collections::HashMap;
use std::str;
// use std::time::Duration;
// use std::time::Instant;
use warp::http::status::StatusCode;

pub fn get_project_version() -> String {
    serde_json::json!({
        "cargo": std::env!("CARGO_PKG_VERSION"),
        "git_describe": std::option_env!("GIT_DESCRIBE"),
        "build_profile": std::option_env!("PROFILE"),
    })
    .to_string()
}

pub fn get_item_tx(tx: &Tx, schema: &Schema, id: &str) -> Result<Vec<Value>> {
    info!("Getting item {}", id);
    let search_query = Search {
        id: Some(id.to_string()),
        _type: None,
        date_server_modified_gte: None,
        date_server_modified_lt: None,
        deleted: None,
        sort_order: SortOrder::Asc,
        limit: 1,
        return_properties: vec![],
        forward_edges: None,
        backward_edges: None,
        other_properties: Default::default(),
    };
    let result = search(tx, schema, search_query)?;
    Ok(result)
}

/// Generate a new random item id.
/// This implementation chooses to generate 32 random hex characters.
pub fn new_random_item_id() -> String {
    new_random_string(32)
}
pub fn new_random_string(length: usize) -> String {
    let mut rng = rand::thread_rng();
    (0..length)
        .map(|_| {
            let idx = rng.gen_range(0..HEX_CHARACTERS.len());
            HEX_CHARACTERS[idx] as char
        })
        .collect()
}
const HEX_CHARACTERS: &[u8] = b"0123456789abcdef";

pub fn create_item_tx(
    tx: &Tx,
    schema: &mut Schema,
    item: CreateItem,
    pod_owner: &str,
    cli: &CliOptions,
    database_key: &DatabaseKey,
) -> Result<String> {
    let id: String = if let Some(id) = &item.id {
        schema::validate_create_item_id(id)?;
        id.to_string()
    } else {
        new_random_item_id()
    };
    let time_now = schema::utc_millis();
    let schema_addition = triggers::add_item_as_schema_opt(tx, schema, &item)?;
    if let SchemaAdditionChange::OldSchema { old_id } = schema_addition {
        Ok(old_id)
    } else {
        let rowid = database_api::insert_item_base(
            tx,
            &id,
            &item._type,
            item.date_created.unwrap_or(time_now),
            item.date_modified.unwrap_or(time_now),
            time_now,
            item.deleted,
        )?;
        for (prop_name, prop_value) in &item.fields {
            insert_property(tx, schema, rowid, prop_name, prop_value)?;
        }
        triggers::trigger_after_item_create(
            tx,
            schema,
            rowid,
            &id,
            &item,
            pod_owner,
            cli,
            database_key,
        )?;
        Ok(id)
    }
}

pub fn update_item_tx(
    tx: &Tx,
    schema: &Schema,
    id: &str,
    fields: HashMap<String, Value>,
) -> Result<()> {
    _update_item_tx(tx, schema, id, fields, None)
}

pub fn _update_item_tx(
    tx: &Tx,
    schema: &Schema,
    id: &str,
    mut fields: HashMap<String, Value>,
    skip_from_readonly: Option<Vec<String>>,
) -> Result<()> {
    log::debug!("Updating item {}", id);
    fields.remove("type");
    fields.remove("dateCreated");

    let readonly_properties: Vec<String> = vec!["containerId".to_string()];

    // if property is readonly but also in skip_from_readonly: still update. This can only be done by the server
    // e.g. when setting the pluginrun.containerId
    for readonly in readonly_properties.into_iter() {
        if let Some(_remove_from_readonly) = &skip_from_readonly {
            if !_remove_from_readonly.contains(&readonly) {
                fields.remove(&readonly);
            }
        } else {
            fields.remove(&readonly);
        }
    }

    let time_now = schema::utc_millis();
    let date_modified = if let Some(dm) = fields.remove("dateModified") {
        if let Some(dm) = dm.as_i64() {
            dm
        } else {
            return Err(Error {
                code: StatusCode::BAD_REQUEST,
                msg: format!("Cannot parse dateModified {} to i64", dm),
            });
        }
    } else {
        time_now
    };
    fields.remove("dateServerModified");
    let deleted = if let Some(d) = fields.remove("deleted") {
        match d {
            Value::Null => None,
            Value::Bool(d) => Some(d),
            _ => {
                return Err(Error {
                    code: StatusCode::BAD_REQUEST,
                    msg: format!("Failed to parse deleted {} to bool", d),
                })
            }
        }
    } else {
        None
    };
    for k in fields.keys() {
        validate_property_name(k)?;
    }
    let rowid = database_api::get_item_rowid(tx, id)?.ok_or_else(|| Error {
        code: StatusCode::NOT_FOUND,
        msg: format!("Item with id {} not found", id),
    })?;
    database_api::update_item_base(tx, rowid, date_modified, time_now, deleted)?;
    for (k, v) in &fields {
        insert_property(tx, schema, rowid, k, v)?;
    }
    Ok(())
}

pub fn delete_item_by_id(tx: &Tx, id: &str) -> Result<()> {
    log::debug!("Deleting item {}", id);
    let rowid = database_api::get_item_rowid(tx, id)?.ok_or_else(|| Error {
        code: StatusCode::NOT_FOUND,
        msg: format!("Item with id {} not found", id),
    })?;
    delete_item_by_rowid(tx, rowid)
}

pub fn delete_item_by_rowid(tx: &Tx, rowid: Rowid) -> Result<()> {
    let time_now = schema::utc_millis();
    database_api::update_item_base(tx, rowid, time_now, time_now, Some(true))
}

pub fn bulk_tx(
    tx: &Tx,
    schema: &mut Schema,
    bulk: Bulk,
    pod_owner: &str,
    cli: &CliOptions,
    database_key: &DatabaseKey,
) -> Result<Value> {
    info!(
        "Performing bulk action with {} new items, {} updated items, {} deleted items, {} created edges",
        bulk.create_items.len(),
        bulk.update_items.len(),
        bulk.delete_items.len(),
        bulk.create_edges.len(),
    );
    let mut created_items = Vec::new();
    for item in bulk.create_items {
        let id = create_item_tx(tx, schema, item, pod_owner, cli, database_key)?;
        created_items.push(id);
    }
    for item in bulk.update_items {
        update_item_tx(tx, schema, &item.id, item.fields)?;
    }
    for item_id in bulk.delete_items {
        delete_item_by_id(tx, &item_id)?;
    }
    let mut created_edges = Vec::new();
    for item_id in bulk.create_edges {
        let id = create_edge(tx, schema, item_id)?;
        created_edges.push(id);
    }
    let mut search_results = Vec::new();
    for query in bulk.search {
        let result = search(tx, schema, query)?;
        search_results.push(result);
    }
    let result = serde_json::json!({
        "createItems": created_items,
        "createEdges": created_edges,
        "search": search_results,
    });
    Ok(result)
}

pub fn create_edge(tx: &Tx, schema: &mut Schema, query: CreateEdge) -> Result<String> {
    let CreateEdge {
        source,
        target,
        name,
        self_id,
    } = query;
    let date = schema::utc_millis();

    let (self_rowid, self_id) = if let Some(id) = self_id {
        let self_rowid = database_api::get_item_rowid(tx, &id)?.ok_or_else(|| Error {
            code: StatusCode::BAD_REQUEST,
            msg: format!("Failed to create edge with _self: {}", id),
        })?;
        (self_rowid, id)
    } else {
        let self_id = new_random_item_id();
        let self_rowid =
            database_api::insert_item_base(tx, &self_id, "Edge", date, date, date, false)?;
        (self_rowid, self_id)
    };

    let source: Rowid = database_api::get_item_rowid(tx, &source)?.ok_or_else(|| Error {
        code: StatusCode::NOT_FOUND,
        msg: format!("Edge source not found: {}", source),
    })?;
    let target: Rowid = database_api::get_item_rowid(tx, &target)?.ok_or_else(|| Error {
        code: StatusCode::NOT_FOUND,
        msg: format!("Edge target not found: {}", target),
    })?;
    insert_edge(tx, schema, self_rowid, source, &name, target)?;
    database_api::update_item_date_server_modified(tx, source, date)?;
    Ok(self_id)
}

pub fn delete_edge_by_source_target(tx: &Tx, query: DeleteEdgeBySourceTarget) -> Result<()> {
    log::debug!(
        "Deleting edge with source {}, target {}",
        query.source,
        query.target
    );
    let source = database_api::get_item_rowid(tx, &query.source)?.ok_or_else(|| Error {
        code: StatusCode::BAD_REQUEST,
        msg: format!(
            "Edge deletion's _source item ID not found {}",
            &query.source
        ),
    })?;
    let target = database_api::get_item_rowid(tx, &query.target)?.ok_or_else(|| Error {
        code: StatusCode::BAD_REQUEST,
        msg: format!(
            "Edge deletion's _target item ID not found {}",
            &query.target
        ),
    })?;
    let self_rowid = database_api::get_edge_self_by_source_target(tx, source, target)?;
    let self_rowid = if let Some(self_rowid) = self_rowid {
        self_rowid
    } else {
        return Err(Error {
            code: StatusCode::NOT_FOUND,
            msg: format!(
                "Cannot delete edge, edge with source {} ({}) and target {} ({}) not found",
                query.source, source, query.target, target
            ),
        });
    };
    database_api::delete_edge(tx, self_rowid)?;
    delete_item_by_rowid(tx, self_rowid)?;
    Ok(())
}

pub fn edge_pointer_to_json(tx: &Tx, schema: &Schema, edge: &EdgePointer) -> Result<Value> {
    let target = database_api::get_item_base(tx, edge.item)?;
    let target = target.ok_or_else(|| Error {
        code: StatusCode::INTERNAL_SERVER_ERROR,
        msg: format!("Edge connects to an nonexisting item.rowid {}", edge.item),
    })?;
    let target_json = Value::Object(item_base_to_json(tx, target, schema)?);
    let edge_item = database_api::get_item_base(tx, edge.rowid)?.ok_or_else(|| Error {
        code: StatusCode::INTERNAL_SERVER_ERROR,
        msg: format!("Edge does not have an item base, rowid: {}", edge.item),
    })?;
    let mut edge_item = item_base_to_json(tx, edge_item, schema)?;
    edge_item.insert("_edge".to_string(), serde_json::json!(edge.name));
    edge_item.insert("_item".to_string(), serde_json::json!(target_json));
    Ok(serde_json::json!(edge_item))
}
pub fn edge_pointers_to_json(
    tx: &Tx,
    schema: &Schema,
    half_edges: &[EdgePointer],
) -> Result<Vec<Value>> {
    let mut result = Vec::new();
    for edge in half_edges {
        result.push(edge_pointer_to_json(tx, schema, edge)?);
    }
    Ok(result)
}

pub fn get_edges(tx: &Tx, query: GetEdges, schema: &Schema) -> Result<Vec<Value>> {
    let root_item = database_api::get_item_rowid(tx, &query.item)?.ok_or_else(|| Error {
        code: StatusCode::NOT_FOUND,
        msg: format!("Cannot find item id {}", query.item),
    })?;
    let edges = if query.direction == EdgeDirection::Outgoing {
        database_api::get_outgoing_edges(tx, root_item)?
    } else {
        database_api::get_incoming_edges(tx, root_item)?
    };
    let mut result = Vec::new();
    for edge in edges {
        let base = database_api::get_item_base(tx, edge.item)?;
        let base = base.ok_or_else(|| Error {
            code: StatusCode::INTERNAL_SERVER_ERROR,
            msg: format!("Edge connects to an nonexisting item.rowid {}", edge.item),
        })?;
        if query.expand_items {
            let item_json = Value::Object(item_base_to_json(tx, base, schema)?);
            result.push(serde_json::json!({
                "name": edge.name,
                "item": item_json,
            }));
        } else {
            result.push(serde_json::json!({
                "name": edge.name,
                "item": { "id": base.id },
            }));
        }
    }
    Ok(result)
}

pub fn search(tx: &Tx, schema: &Schema, query: Search) -> Result<Vec<Value>> {
    info!("Searching by fields {:?}", query);
    if !query.other_properties.is_empty() {
        log::info!(
            "Doing a slow search request using non-base properties {:?}",
            query.other_properties.keys()
        );
    }
    let database_search = DatabaseSearch {
        rowid: None,
        id: query.id.as_deref(),
        _type: query._type.as_deref(),
        date_server_modified_gte: query.date_server_modified_gte,
        date_server_modified_lt: query.date_server_modified_lt,
        deleted: query.deleted,
        sort_order: query.sort_order.clone(),
        _limit: query.limit,
    };
    let items = database_api::search_items_with_properties(
        tx,
        &database_search,
        Some(schema),
        Some(&query.other_properties),
    )?;
    let mut result = Vec::new();
    for item in items {
        let rowid = item.rowid;
        let object_map = item_base_to_json(tx, item, schema)?;
        let mut object_map = if query.return_properties.is_empty() {
            object_map
        } else {
            object_map
                .into_iter()
                .filter(|(prop, _value)| query.return_properties.contains(prop))
                .collect()
        };
        add_item_edge_properties(tx, &mut object_map, rowid)?;
        if query.forward_edges.is_some() {
            let edges = get_outgoing_edges(tx, rowid)?;
            let edges = edge_pointers_to_json(tx, schema, &edges)?;
            object_map.insert("[[edges]]".to_string(), Value::Array(edges));
        }
        if query.backward_edges.is_some() {
            let edges = get_incoming_edges(tx, rowid)?;
            let edges = edge_pointers_to_json(tx, schema, &edges)?;
            object_map.insert("~[[edges]]".to_string(), Value::Array(edges));
        }
        result.push(Value::Object(object_map));
        // }
    }
    Ok(result)
}

fn gql_search_recursive(
    tx: &Tx,
    query_ast: &QueryASTNode,
    schema: &Schema,
    item_rowids: Option<Vec<Rowid>>,
    level: u32,
) -> Result<Vec<Value>> {
    let sort_property = query_ast
        .arguments
        .sort_property
        .to_owned()
        .unwrap_or_else(|| "dateServerModified".to_owned());
    let sort_order = query_ast
        .arguments
        .sort_order
        .to_owned()
        .unwrap_or_default();

    let query = GQLSearchArgs {
        _type: query_ast.item_type.to_owned(),
        properties: query_ast.properties.to_owned(),
        rowids: item_rowids,
        sort_property,
        sort_order,
        limit: query_ast.arguments.limit,
        offset: query_ast.arguments.offset,
        filters: query_ast.arguments.filters.to_owned(),
    };

    // let now = Instant::now();

    let items = database_api::search_gql_properties(tx, &query, schema)?;

    let mut result = Vec::new();
    // if level == 0 {
    // let elapsed = now.elapsed();
    // println!("Elapsed: {:.2?}", elapsed);
    // }

    // let now = Instant::now();
    // let mut total_edges = Duration::new(0, 0);
    // let mut total_edges_items = Duration::new(0, 0);
    for item in items.into_iter() {
        let mut item_json = item.properties;
        for (edge_name, edge_item_ast) in &query_ast.edges {
            // let now_edges = Instant::now();
            let edges = if edge_name.contains('~') {
                let normalized = &edge_name[1..edge_name.len()];
                database_api::get_edges(tx, item.rowid, Some(&normalized.to_string()), true)?
            } else {
                database_api::get_edges(tx, item.rowid, Some(edge_name), false)?
            };
            // if level == 0 {
            //     let elapsed = now_edges.elapsed();
            //     total_edges += elapsed;
            // }
            let target_ids: Option<Vec<Rowid>> = Some(edges.iter().map(|e| e.target).collect());
            // let now_edge_items = Instant::now();
            // if target_ids.is_some() && &target_ids.unwrap().len() > &0 {
            let edge_jsons = if let Some(_target_ids) = target_ids {
                if !_target_ids.is_empty() {
                    gql_search_recursive(tx, edge_item_ast, schema, Some(_target_ids), level + 1)?
                } else {
                    Vec::new()
                }
            } else {
                Vec::new()
            };

            item_json.insert(edge_name.to_owned(), edge_jsons.into());
            // if level == 0 {
            //     let elapsed = now_edge_items.elapsed();
            //     total_edges_items += elapsed;
            // }
        }
        result.push(item_json.into());
    }
    // if level == 0 {
    // println!("Elapsed edges: {:.2?}", total_edges);
    // println!("Elapsed edges items: {:.2?}", total_edges_items);
    // let elapsed = now.elapsed();
    // println!("Total Elapsed: {:.2?}", elapsed);
    // }
    Ok(result)
}

pub fn graphql(tx: &Tx, schema: &Schema, query: String) -> Result<HashMap<String, Vec<Value>>> {
    let query_ast: QueryASTNode = graphql_utils::parse_graphql_query(&query).unwrap();
    let search_result = gql_search_recursive(tx, &query_ast, schema, None, 0)?;
    let mut result = HashMap::new();
    result.insert("data".to_owned(), search_result);
    Ok(result)
}

#[cfg(test)]
pub mod tests {
    use crate::api_model::CreateItem;
    use crate::command_line_interface;
    use crate::database_api;
    use crate::database_api::new_conn;
    use crate::error::Result;
    use crate::internal_api;
    use crate::internal_api::*;
    use crate::plugin_auth_crypto::DatabaseKey;
    use crate::schema::Schema;
    use serde_json::json;
    use warp::hyper::StatusCode;

    #[test]
    fn test_schema_checking() -> Result<()> {
        let mut conn = new_conn();
        let tx = conn.transaction().unwrap();
        let database_key = DatabaseKey::from("".to_string()).unwrap();
        let cli = command_line_interface::tests::test_cli();

        // first try to insert the Person without Schema
        let item_json = json!({
            "type": "Person",
            "age": 20,
        });
        let item_struct: CreateItem = serde_json::from_value(item_json).unwrap();
        let result = internal_api::create_item_tx(
            &tx,
            &mut database_api::get_schema(&tx)?,
            item_struct.clone(),
            "",
            &cli,
            &database_key,
        );
        assert_eq!(result.unwrap_err().code, StatusCode::BAD_REQUEST);

        // Then insert the Schema, it should succeed
        let schema_json = json!({
            "type": "ItemPropertySchema",
            "itemType": "Person",
            "propertyName": "age",
            "valueType": "Integer",
        });
        let schema_struct: CreateItem = serde_json::from_value(schema_json).unwrap();
        let result = internal_api::create_item_tx(
            &tx,
            &mut database_api::get_schema(&tx)?,
            schema_struct,
            "",
            &cli,
            &database_key,
        );
        assert!(result.is_ok());

        // Now insert the Person with already in place
        let result = internal_api::create_item_tx(
            &tx,
            &mut database_api::get_schema(&tx)?,
            item_struct,
            "",
            &cli,
            &database_key,
        );
        assert!(result.is_ok());

        Ok(())
    }

    #[test]
    fn test_item_insert_schema() {
        let mut conn = new_conn();
        let cli = command_line_interface::tests::test_cli();
        let database_key = DatabaseKey::from("".to_string()).unwrap();

        let tx = conn.transaction().unwrap();
        let mut minimal_schema = database_api::get_schema(&tx).unwrap();

        let json = json!({
            "type": "ItemPropertySchema",
            "itemType": "Person",
            "propertyName": "age",
            "valueType": "Integer",
        });

        {
            // insert "age" successfully
            let create_item: CreateItem = serde_json::from_value(json.clone()).unwrap();
            internal_api::create_item_tx(
                &tx,
                &mut minimal_schema,
                create_item,
                "",
                &cli,
                &database_key,
            )
            .unwrap();
        }

        {
            let json = json!({
                "type": "ItemPropertySchema",
                "itemType": "Person",
                "propertyName": "dateCreated",
                "valueType": "Bool",
            });
            let create_item: CreateItem = serde_json::from_value(json).unwrap();
            let expected_error = "Schema for property dateCreated is already defined to valueType DateTime, cannot override to Bool";
            let result = internal_api::create_item_tx(
                &tx,
                &mut minimal_schema,
                create_item,
                "",
                &cli,
                &database_key,
            );
            assert!(result.is_err());
            assert!(result.unwrap_err().msg.contains(expected_error));
        }

        let mut bad_empty_schema = Schema::empty();
        let create_item: CreateItem = serde_json::from_value(json).unwrap();
        let result = internal_api::create_item_tx(
            &tx,
            &mut bad_empty_schema,
            create_item,
            "",
            &cli,
            &database_key,
        )
        .unwrap_err();
        assert_eq!(result.code, StatusCode::BAD_REQUEST);
        assert!(result.msg.contains("not defined in Schema"));

        tx.commit().unwrap();
    }

    #[test]
    fn test_search() {
        let mut conn = new_conn();
        let cli = command_line_interface::tests::test_cli();
        let db_key = DatabaseKey::from("".to_string()).unwrap();
        let tx = conn.transaction().unwrap();
        let mut schema = database_api::get_schema(&tx).unwrap();

        let person = {
            let person = json!({
                "type": "Person",
            });
            let person: CreateItem = serde_json::from_value(person).unwrap();
            create_item_tx(&tx, &mut schema, person, "", &cli, &db_key).unwrap()
        };
        {
            let json = json!({
                "id": person,
            });
            let parsed = serde_json::from_value(json).unwrap();
            let result = search(&tx, &schema, parsed).unwrap();
            assert_eq!(result.len(), 1);
            let result = result.into_iter().next().unwrap();
            let result = match result {
                Value::Object(r) => r,
                _ => panic!("Result is not JSON Object"),
            };
            assert!(result.len() > 3);
        };
        {
            let json = json!({
                "id": person,
                "_returnProperties": ["id"]
            });
            let parsed = serde_json::from_value(json).unwrap();
            let result = search(&tx, &schema, parsed).unwrap();
            assert_eq!(result.len(), 1);
            let result = result.into_iter().next().unwrap();
            let result = match result {
                Value::Object(r) => r,
                _ => panic!("Result is not JSON Object"),
            };
            assert_eq!(result.len(), 1);
            assert!(result.get("id").is_some());
        };
    }

    fn prepare_graphql_dataset(
        tx: &Tx,
        pod_owner: &str,
        cli: CliOptions,
        database_key: DatabaseKey,
    ) {
        // insert age prop to Person
        let schema_json = json!({
            "type": "ItemPropertySchema",
            "itemType": "Person",
            "propertyName": "age",
            "valueType": "Integer",
        });
        let schema_struct: CreateItem = serde_json::from_value(schema_json).unwrap();
        let mut schema = database_api::get_schema(tx).unwrap();
        create_item_tx(tx, &mut schema, schema_struct, "", &cli, &database_key).unwrap();

        let mut schema = database_api::get_schema(tx).unwrap();
        // create level 1
        for i in 0..1000 {
            let person = json!({
                "type": "Person1",
                "id": "00000".to_owned() + &i.to_string(),
                "age": i
            });
            let person: CreateItem = serde_json::from_value(person).unwrap();
            create_item_tx(tx, &mut schema, person, pod_owner, &cli, &database_key).unwrap();
        }

        // create level 2
        for i in 1000..2000 {
            let person = json!({
                "type": "Person2",
                "id": "00000".to_owned() + &i.to_string(),
                "age": i
            });
            let person: CreateItem = serde_json::from_value(person).unwrap();
            create_item_tx(tx, &mut schema, person, pod_owner, &cli, &database_key).unwrap();
        }

        // create level 3
        for i in 2000..12000 {
            let person = json!({
                "type": "Person3",
                "id": "00000".to_owned() + &i.to_string()
            });
            let person: CreateItem = serde_json::from_value(person).unwrap();
            create_item_tx(tx, &mut schema, person, pod_owner, &cli, &database_key).unwrap();
        }

        // create schema for connections
        let schema_json = json!({
            "type": "ItemEdgeSchema",
            "edgeName": "friend1",
            "sourceType": "Person1",
            "targetType": "Person2",
        });
        let schema_struct: CreateItem = serde_json::from_value(schema_json).unwrap();
        let result = create_item_tx(
            tx,
            &mut database_api::get_schema(tx).unwrap(),
            schema_struct,
            pod_owner,
            &cli,
            &database_key,
        );
        assert!(result.is_ok());

        let schema_json = json!({
            "type": "ItemEdgeSchema",
            "edgeName": "friend2",
            "sourceType": "Person2",
            "targetType": "Person3",
        });
        let schema_struct: CreateItem = serde_json::from_value(schema_json).unwrap();
        let result = create_item_tx(
            tx,
            &mut database_api::get_schema(tx).unwrap(),
            schema_struct,
            pod_owner,
            &cli,
            &database_key,
        );
        assert!(result.is_ok());

        // Get updated schema
        let mut schema = database_api::get_schema(tx).unwrap();

        // create edges level 1 -> level 2
        for i in 0..100 {
            let _edge = {
                let json = json!({
                    "_source": "00000".to_owned() + &i.to_string(),
                    "_target": "00000".to_owned() + &(i+1000).to_string(),
                    "_name": "friend1",
                });
                let parsed = serde_json::from_value(json).unwrap();
                create_edge(tx, &mut schema, parsed).unwrap()
            };
        }
        // create edges level 2 -> level 3
        for i in 0..1000 {
            let _edge = {
                let json = json!({
                    "_source": "00000".to_owned() + &((i/10)+1000).to_string(),
                    "_target": "00000".to_owned() + &(i+2000).to_string(),
                    "_name": "friend2",
                });
                let parsed = serde_json::from_value(json).unwrap();
                create_edge(tx, &mut schema, parsed).unwrap()
            };
        }
    }

    #[test]
    fn test_graphql() {
        let mut conn = new_conn();

        // let mut conn = rusqlite::Connection::open("test.sqlite").unwrap();
        // database_migrate_refinery::embedded::migrations::runner()
        //     .run(&mut conn)
        //     .expect("Failed to run refinery migrations");

        let tx = conn.transaction().unwrap();
        let database_key = DatabaseKey::from("".to_string()).unwrap();
        let cli = command_line_interface::default_cli();
        let pod_owner = "";

        prepare_graphql_dataset(&tx, pod_owner, cli, database_key);

        let schema = database_api::get_schema(&tx).unwrap();

        let query = "
        query {
            Person1 {
               id 
               age
               friend1 {
                   id
                   age
                   friend2 {
                       id
                       age
                   }
               }
            }
        }"
        .to_owned();

        let res = graphql(&tx, &schema, query);
        let data = res.unwrap();
        let data = data.get("data").unwrap();
        // assert_eq!(data.len(), 2000);

        let items_with_friends = data
            .iter()
            .filter_map(|x| x.get("friend1"))
            .filter(|x| !x.as_array().unwrap().is_empty());

        // print!("{:}", serde_json::to_string_pretty(&json!(data)).unwrap());
        assert_eq!(
            items_with_friends.clone().count(),
            100,
            "Expected 100 friends"
        );

        // check if 1st level prop is fetched
        let age: i32 = serde_json::from_value(data[0]["age"].clone()).unwrap_or(0);
        assert!(age > 0, "Cannot get root property: age");

        // // check 2nd level prop
        for friends in items_with_friends {
            let friend_age: i32 = serde_json::from_value(friends[0]["age"].clone()).unwrap_or(0);
            assert!(friend_age > 0, "Cannot get edge item property: age");
        }

        // test limit and offset
        let query = "
        query {
            Person1 (limit: 10, offset: 5) {
               id 
               friend1 {
                    id
                }
            }
        }"
        .to_owned();
        // print!("{:}", serde_json::to_string_pretty(&json!(data)).unwrap());

        let res = graphql(&tx, &schema, query);
        let data = res.unwrap();
        let data = data.get("data").unwrap();
        assert_eq!(data.len(), 10);
    }

    #[test]
    fn test_graphql_filters() {
        let mut conn = new_conn();

        // let mut conn = rusqlite::Connection::open("test.sqlite").unwrap();
        // database_migrate_refinery::embedded::migrations::runner()
        //     .run(&mut conn)
        //     .expect("Failed to run refinery migrations");

        let tx = conn.transaction().unwrap();
        let database_key = DatabaseKey::from("".to_string()).unwrap();
        let cli = command_line_interface::default_cli();
        let pod_owner = "";

        prepare_graphql_dataset(&tx, pod_owner, cli, database_key);

        let schema = database_api::get_schema(&tx).unwrap();

        let query = "
        query {
            Person1 {
               id 
               friend1 {
                   id
                   friend2 {
                       id
                   }
               }
            }
        }"
        .to_owned();

        let res = graphql(&tx, &schema, query);
        let data = res.unwrap();
        let data = data.get("data").unwrap();
        // assert_eq!(data.len(), 2000);

        let items_with_friends = data
            .iter()
            .filter_map(|x| x.get("friend1"))
            .filter(|x| !x.as_array().unwrap().is_empty());

        // print!("{:}", serde_json::to_string_pretty(&json!(data)).unwrap());
        assert_eq!(items_with_friends.count(), 100);

        let query = "
        query {
            Person1 (filter: {id: {ne: \"000000\"}}) {
               id 
               friend1 {
                   id
               }
            }
        }"
        .to_owned();
        let res = graphql(&tx, &schema, query);
        let data = res.unwrap();
        let data = data.get("data").unwrap();
        assert_eq!(data.len(), 999);
    }

    #[test]
    /// Test creating an item that has type="Edge"
    fn test_create_edge_item() {
        let mut conn = new_conn();
        let cli = command_line_interface::tests::test_cli();
        let db_key = DatabaseKey::from("".to_string()).unwrap();
        let tx = conn.transaction().unwrap();
        let mut schema = database_api::get_schema(&tx).unwrap();
        let edge_item = json!({
            "type": "Edge",
        });
        let edge_item: CreateItem = serde_json::from_value(edge_item).unwrap();
        // Creating raw "Edge" items should be successful.
        // (It is also the way how Memri clients push edge information to the Pod.)
        create_item_tx(&tx, &mut schema, edge_item, "", &cli, &db_key).unwrap();
    }

    #[test]
    fn test_edge_schema() {
        let mut conn = new_conn();
        let tx = conn.transaction().unwrap();
        let database_key = DatabaseKey::from("".to_string()).unwrap();
        let cli = command_line_interface::tests::test_cli();

        // Then insert the Schema, it should succeed
        let schema_json = json!({
          "type": "ItemEdgeSchema",
          "edgeName": "friend",
          "sourceType": "Person",
          "targetType": "Person",
        });
        let schema_struct: CreateItem = serde_json::from_value(schema_json).unwrap();
        let result = internal_api::create_item_tx(
            &tx,
            &mut database_api::get_schema(&tx).unwrap(),
            schema_struct,
            "",
            &cli,
            &database_key,
        );
        assert!(result.is_ok());
    }

    #[test]
    fn test_edge_search() {
        let mut conn = new_conn();
        let cli = command_line_interface::tests::test_cli();
        let db_key = DatabaseKey::from("".to_string()).unwrap();
        let tx = conn.transaction().unwrap();

        let schema_json = json!({
          "type": "ItemEdgeSchema",
          "edgeName": "friend",
          "sourceType": "Person",
          "targetType": "Person",
        });
        let schema_struct: CreateItem = serde_json::from_value(schema_json).unwrap();
        let result = internal_api::create_item_tx(
            &tx,
            &mut database_api::get_schema(&tx).unwrap(),
            schema_struct,
            "",
            &cli,
            &db_key,
        );
        assert!(result.is_ok());

        let mut schema = database_api::get_schema(&tx).unwrap();

        let person1 = {
            let person = json!({
                "type": "Person",
            });
            let person: CreateItem = serde_json::from_value(person).unwrap();
            create_item_tx(&tx, &mut schema, person, "", &cli, &db_key).unwrap()
        };
        let person2 = {
            let person = json!({
                "type": "Person",
            });
            let person: CreateItem = serde_json::from_value(person).unwrap();
            create_item_tx(&tx, &mut schema, person, "", &cli, &db_key).unwrap()
        };

        let _edge = {
            let json = json!({
                "_source": person1,
                "_target": person2,
                "_name": "friend",
            });
            let parsed = serde_json::from_value(json).unwrap();
            create_edge(&tx, &mut schema, parsed).unwrap()
        };

        let result = {
            let json = json!({
                "id": person1,
                "[[edges]]": {},
            });
            let parsed = serde_json::from_value(json).unwrap();
            search(&tx, &schema, parsed).unwrap()
        };
        assert_eq!(result.len(), 1);
        let result = result.into_iter().next().unwrap();
        let result = match result {
            Value::Object(r) => r,
            _ => panic!("Result is not JSON Object"),
        };

        // Example:
        // {
        //   "[[edges]]": [
        //     {
        //       "_edge": "friend",
        //       "_item": {
        //         "dateCreated": 1624363823546,
        //         "dateModified": 1624363823546,
        //         "dateServerModified": 1624363823546,
        //         "deleted": false,
        //         "id": "1539da9dfc14a0467694e79730135df6",
        //         "type": "Person"
        //       },
        //       "dateCreated": 1624363823546,
        //       "dateModified": 1624363823546,
        //       "dateServerModified": 1624363823546,
        //       "deleted": false,
        //       "id": "4ea088d17a670cbb26f4154bce809950",
        //       "type": "Edge"
        //     }
        //   ],
        //   "dateCreated": 1624363823545,
        //   "dateModified": 1624363823545,
        //   "dateServerModified": 1624363823545,
        //   "deleted": false,
        //   "id": "59dd635884bff1de1c8447ac3763543b",
        //   "type": "Person"
        // }

        let result_id = result.get("id").unwrap().as_str().unwrap();
        assert_eq!(result_id, person1);

        let result_edge = result.get("[[edges]]").unwrap().as_array().unwrap();
        let result_edge = result_edge.iter().next().unwrap();

        assert_eq!(
            result_edge.get("_edge").unwrap().as_str().unwrap(),
            "friend"
        );

        let result_edge_item = result_edge.get("_item").unwrap();
        let result_edge_item = match result_edge_item {
            Value::Object(i) => i,
            _ => panic!("edge item is not an JSON Object"),
        };
        assert_eq!(
            result_edge_item.get("id").unwrap().as_str().unwrap(),
            person2
        );
    }

    #[test]
    fn test_edge_delete() {
        let mut conn = new_conn();
        let cli = command_line_interface::tests::test_cli();
        let db_key = DatabaseKey::from("".to_string()).unwrap();
        let tx = conn.transaction().unwrap();

        let schema_json = json!({
          "type": "ItemEdgeSchema",
          "edgeName": "friend",
          "sourceType": "Person",
          "targetType": "Person",
        });
        let schema_struct: CreateItem = serde_json::from_value(schema_json).unwrap();
        let result = internal_api::create_item_tx(
            &tx,
            &mut database_api::get_schema(&tx).unwrap(),
            schema_struct,
            "",
            &cli,
            &db_key,
        );
        assert!(result.is_ok());

        let mut schema = database_api::get_schema(&tx).unwrap();

        let person1 = {
            let person = json!({
                "type": "Person",
            });
            let person: CreateItem = serde_json::from_value(person).unwrap();
            create_item_tx(&tx, &mut schema, person, "", &cli, &db_key).unwrap()
        };
        let person2 = {
            let person = json!({
                "type": "Person",
            });
            let person: CreateItem = serde_json::from_value(person).unwrap();
            create_item_tx(&tx, &mut schema, person, "", &cli, &db_key).unwrap()
        };
        let _edge = {
            let json = json!({
                "_source": person1,
                "_target": person2,
                "_name": "friend",
            });
            let parsed = serde_json::from_value(json).unwrap();
            create_edge(&tx, &mut schema, parsed).unwrap()
        };

        let result = {
            let json = json!({
                "id": person1,
                "[[edges]]": {},
            });
            let parsed = serde_json::from_value(json).unwrap();
            search(&tx, &schema, parsed).unwrap()
        };
        assert_eq!(result.len(), 1);

        {
            let json = json!({
                "_source": person1,
                "_target": person2,
            });
            let parsed = serde_json::from_value(json).unwrap();
            delete_edge_by_source_target(&tx, parsed).unwrap()
        };
        let result = {
            let json = json!({
                "id": person1,
                "[[edges]]": {},
            });
            let parsed = serde_json::from_value(json).unwrap();
            search(&tx, &schema, parsed).unwrap()
        };
        assert_eq!(result.len(), 1);
        let result = result.into_iter().next().unwrap();
        let result = match result {
            Value::Object(r) => r,
            _ => panic!("Result is not JSON Object"),
        };
        let result_edges = result.get("[[edges]]").unwrap().as_array().unwrap();
        assert!(
            result_edges.is_empty(),
            "There should be 0 edges after the deletion of the edge"
        );
    }
}
