//
// Pod triggers that should "run" on particular DB/data changes (e.g. item insertion)
//

use crate::api_model::CreateItem;
use crate::command_line_interface::CliOptions;
use crate::database_api;
use crate::database_api::Rowid;
use crate::database_utils::get_item_from_rowid;
use crate::error::Error;
use crate::error::ErrorContext;
use crate::error::Result;
use crate::plugin_auth_crypto::DatabaseKey;
use crate::plugin_run;
use crate::schema::Schema;
use crate::schema::SchemaPropertyType;
use crate::triggers::SchemaAdditionChange::*;
use rusqlite::Transaction as Tx;
use serde::Deserialize;
use serde::Serialize;
use warp::http::StatusCode;

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct SchemaItem {
    pub item_type: String,
    pub property_name: String,
    pub value_type: SchemaPropertyType,
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct PluginRunItem {
    pub container_image: String,
    pub target_item_id: String,
}

#[derive(Debug, Eq, PartialEq)]
pub enum SchemaAdditionChange {
    NotASchema,
    NewSchemaAdded,
    OldSchema { old_id: String },
}

/// If an item is a Schema, add it to the schema.
/// Fail if an incompatible schema is attempted to be inserted.
pub fn add_item_as_schema_opt(
    tx: &Tx,
    schema: &mut Schema,
    item: &CreateItem,
) -> Result<SchemaAdditionChange> {
    if item._type == "ItemPropertySchema" {
        // We'll do something ugly here.
        // We'll convert the item into JSON, and then back into the desired type
        // for type check and parsing. This is easier code-wise than to do manual conversions.
        // It only triggers for specific, rarely used items. This implementation might change later.
        let json = serde_json::to_value(item)?;
        let parsed: SchemaItem = serde_json::from_value(json)
            .context(|| format!("Parsing of Schema item {:?}, {}:{}", item, file!(), line!()))?;

        // The result of the operation depends on:
        // * the "heterogeneous" property's type
        // * this particular itemType's property type
        // * whether an ID is present in the request
        match (
            schema.property(&parsed.property_name),
            schema.item_property(&parsed.item_type, &parsed.property_name),
            &item.id,
        ) {
            (Some(old_value_type), _, _) if &parsed.value_type != old_value_type => {
                Err(Error {
                    code: StatusCode::BAD_REQUEST,
                    msg: format!(
                        "Schema for property {} is already defined to valueType {}, cannot override to {}",
                        parsed.property_name,
                        old_value_type,
                        parsed.value_type,
                    )
                })
            }
            (_, None, _) | (_, _, Some(_)) => {
                schema.set_type_unchecked(parsed.item_type, parsed.property_name, parsed.value_type);
                Ok(NewSchemaAdded)
            }
            (_, Some(old_value_type), None) if &parsed.value_type == old_value_type => {
                let old_id = database_api::get_schema_item_id_for_property(
                    tx,
                    &parsed.item_type,
                    &parsed.property_name,
                )?;
                Ok(OldSchema { old_id })
            }
            (union_value_type, Some(old_value_type), None) => {
                Err(Error {
                    code: StatusCode::INTERNAL_SERVER_ERROR,
                    msg: format!(
                        "Unexpected failure, {}.{} has valueType {}. However, this Schema should not have been allowed as {} is already defined as {:?}",
                        parsed.item_type,
                        parsed.property_name,
                        old_value_type,
                        parsed.property_name,
                        union_value_type,
                    )
                })
            }
        }
    } else {
        Ok(NotASchema)
    }
}

#[allow(clippy::too_many_arguments)]
pub fn trigger_after_item_create(
    tx: &Tx,
    schema: &Schema,
    source_rowid: Rowid,
    source_id: &str,
    item: &CreateItem,
    pod_owner: &str,
    cli: &CliOptions,
    database_key: &DatabaseKey,
) -> Result<()> {
    if item._type == "PluginRun" {
        let json = get_item_from_rowid(tx, schema, source_rowid)?;
        let parsed: PluginRunItem = serde_json::from_value(json)
            .context(|| format!("Parsing of item {:?}, {}:{}", item, file!(), line!()))?;
        plugin_run::run_plugin_container(
            tx,
            schema,
            parsed.container_image,
            &parsed.target_item_id,
            source_id,
            pod_owner,
            database_key,
            cli,
        )?;
    }
    Ok(())
}

#[cfg(test)]
mod tests {
    use super::add_item_as_schema_opt;
    use crate::api_model::CreateItem;
    use crate::database_api::new_conn;
    use crate::error::Error;
    use crate::error::Result;
    use crate::schema::Schema;
    use crate::schema::SchemaPropertyType;
    use crate::triggers::SchemaAdditionChange;
    use serde_json::json;
    use warp::http::StatusCode;

    #[test]
    fn my_test() -> Result<()> {
        let mut conn = new_conn();
        let tx = conn.transaction().unwrap();

        let mut schema = Schema::empty();
        schema.set_type_unchecked(
            "Person".to_string(),
            "age".to_string(),
            SchemaPropertyType::Integer,
        );

        let json = json!({
            "type": "Something"
        });
        let create_item: CreateItem = serde_json::from_value(json).unwrap();
        let result = add_item_as_schema_opt(&tx, &mut schema, &create_item);
        assert_eq!(result, Ok(SchemaAdditionChange::NotASchema));

        let json = json!({
            "type": "ItemPropertySchema",
            "itemType": "Person",
            "propertyName": "age",
            "valueType": "Integer",
        });
        let create_item: CreateItem = serde_json::from_value(json).unwrap();
        let result = add_item_as_schema_opt(&tx, &mut schema, &create_item);
        assert_eq!(
            result,
            Err(Error {
                code: StatusCode::INTERNAL_SERVER_ERROR,
                msg: "Failed to find property Person.age".to_string()
            })
        );

        let json = json!({
            "type": "ItemPropertySchema",
            "itemType": "Person",
            "propertyName": "agility",
            "valueType": "Integer",
        });
        let create_item: CreateItem = serde_json::from_value(json).unwrap();
        assert_eq!(schema.all_properties_len(), 1);
        let result = add_item_as_schema_opt(&tx, &mut schema, &create_item);
        assert_eq!(schema.all_properties_len(), 2);
        assert_eq!(
            schema.property("agility"),
            Some(SchemaPropertyType::Integer).as_ref()
        );
        assert_eq!(result, Ok(SchemaAdditionChange::NewSchemaAdded));

        let json = json!({
            "type": "ItemPropertySchema",
            "itemType": "Person",
            "propertyName": "age",
            "valueType": "Text",
        });
        let create_item: CreateItem = serde_json::from_value(json).unwrap();
        let result = add_item_as_schema_opt(&tx, &mut schema, &create_item);
        assert!(result.is_err(), "result should be an error {:?}", result);

        Ok(())
    }
}
